/* libmup.c - M.A.R-T>Y. Micro Protocol (MuP) is an endpoint-based request/response RPC
 * protocol for small-ish interlinked systems.

 * The protocol spec can be found at https://hoff.industries/release-mup.html
 * Last updated 06/08/2024

  * NOTE: a description of #define modifiers can be found in libmup.h
  * please report bugs to theworldisyours@mailbox.org

 * This library licensed with BSD 3-Clause:

Copyright 2023-2024 hoff.industries

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>

#ifndef H_LIBMUP
#include "libmup.h"
#endif

const int MUP_LIB_VERSION[3] = {0, 2, 0};
const int MUP_PROTO_VERSION = 0x01;

const char MUP_AUTHOR[] = "hoff.industries";
const char MARTY_FLAVOUR_TEXT[] = "M.A.R-T>Y.";

#ifndef MUP_CFG_BYOC
/* CRC32 table and routine sourced from
 * https://web.mit.edu/freebsd/head/sys/libkern/crc32.c
 * it allows unrestricted use */

// CRC-32/ISO-HDLC poly 0x04C11DB7
const uint32_t crc32_tab[] = {

	0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
	0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
	0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
	0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
	0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
	0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
	0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b, 0x35b5a8fa, 0x42b2986c,
	0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
	0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
	0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
	0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190, 0x01db7106,
	0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
	0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
	0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
	0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
	0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
	0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
	0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
	0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
	0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
	0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
	0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
	0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
	0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
	0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
	0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
	0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
	0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
	0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
	0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
	0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
	0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
	0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
	0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
	0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
	0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
	0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
	0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
	0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
	0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
	0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
	0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
	0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};
#endif

const int PayloadCounts[MUP_MAX_TYPE + 1] = {

	MUP_PAYL_COUNT_UNSET,
	MUP_PAYL_COUNT_INIT,
	MUP_PAYL_COUNT_BAD,
	MUP_PAYL_COUNT_CONTERM,
	MUP_PAYL_COUNT_SONAR,
	MUP_PAYL_COUNT_SEND,
	MUP_PAYL_COUNT_API
};

const int ArgumentBounds[MUP_MAX_TYPE + 1] = {

	MUP_ARG_UNSET_MAX,
	MUP_ARG_INIT_MAX,
	MUP_ARG_BAD_MAX,
	MUP_ARG_CONTERM_MAX,
	MUP_ARG_SONAR_MAX,
	MUP_ARG_SEND_MAX,
	MUP_ARG_API_MAX
};

#ifndef MUP_CFG_BYOC
uint32_t mup_crc32(const void *buf, int size) {

	const uint8_t *p = buf;
	uint32_t crc;

	crc = ~0U;
	while (size--) {

		crc = crc32_tab[(crc ^ *p++) & 0xFF] ^ (crc >> 8);
	}

	return crc ^ ~0U;
}

uint32_t (*mup_calc_crc)(const void *buf, int size) = &mup_crc32;
#else

uint32_t mup_dummy_crc32(const void *buf, int size) {

	return 0x00000000;
}

// WARNING: MAKE SURE TO REGISTER A CRC FUNCTION WITH mup_register_crc_func!!
uint32_t (*mup_calc_crc)(const void *buf, int size) = &mup_dummy_crc32;
#endif

void mup_register_crc_func(uint32_t (*crc_calc_func)(const void *buf, int size)) {

	mup_calc_crc = crc_calc_func;
}

#ifdef MUP_CFG_BYOB
void mup_packet_factory(struct MUPacket *packet, uint8_t *buffer, bool zero) {
#else
void mup_packet_factory(struct MUPacket *packet, bool zero) {
#endif

	packet->version = MUP_PROTO_VERSION;
	packet->type = MUP_TYPE_UNSET;
	packet->arg = MUP_ARG_UNSET_UNSET;
	packet->size = 0;
	packet->crc32 = 0x00000000;

	packet->payloadCount = 0;
	packet->cachedPayload = -1;
	packet->cachedSize = 0;

#ifdef MUP_CFG_BYOB
	packet->data = buffer == NULL ? packet->data : buffer;
#endif

	packet->payloadCache = &packet->data[MUP_HEADER_SIZE_W_CRC];

	for (int i = 0; zero && i < MUP_MAX_SIZE; i++) {

		packet->data[i] = 0x00;
	}
}

void mup_status_factory(struct MUStatus *status) {

	status->size = 0;
	status->error = MUP_ERR_NONE;
	status->event = MUP_EVT_NONE;
	status->doReply = false;
}

void mup_endpoint_factory(struct MUEndpoint *endpoint) {

	endpoint->endpoint = -1;
	endpoint->name = NULL;

	endpoint->min = -1;
	endpoint->max = -1;

	endpoint->isAscii = false;
}

// NOTE: payload and packet len functions act like strlen;
// they do not include the terminator as part of the length.

// WARNING: This can go out of bounds if you don't terminate
// your payloads properly and pass in a buffer smaller than MUP_MAX_PAYLOAD_SIZE
int mup_payload_len(uint8_t *payload) {

	int i;
#ifdef MUP_CFG_CONV_STRINGS
	for (i = 0; payload[i] != MUP_PAYLOAD_TERM && payload[i] != '\0' && i < MUP_MAX_PAYLOAD_SIZE; i++);
#else
	for (i = 0; payload[i] != MUP_PAYLOAD_TERM && i < MUP_MAX_PAYLOAD_SIZE; i++);
#endif
	return i;
}

// NOTE: As long as you always pass the complete packet buffer
// of size MUP_MAX_SIZE this should never go out of bounds.
int mup_packet_len(uint8_t *packet) {
	int i;
	for (i = 0; packet[i] != MUP_PACKET_TERM && i < MUP_MAX_SIZE; i++);
	return i;
}

void mup_register_payload(struct MUPacket *packet, struct MUStatus *status) {

	// NOTE: this should never be OOB in register - only set by library
	uint8_t *startAddr = packet->payloadCache +
		sizeof(uint8_t)*packet->cachedSize;

	int paylen = mup_payload_len(startAddr);

	// payload/packet would not have been terminated properly
	if (paylen >= MUP_MAX_SIZE - 2) {

		status->error = MUP_ERR_BAD_PACK_LEN;
		return;
	}

	packet->payloadCache = startAddr;

	packet->payloadCount++;
	packet->cachedPayload++;
	// paylen incremented to allow for terminator
	packet->cachedSize = ++paylen;

	packet->size += paylen;
	status->size = paylen;
}

void mup_add_payload(struct MUPacket *packet,
		uint8_t *payload, struct MUStatus *status) {

	int paylen;

	// NOTE: this should never be OOB internally - only set by library
	uint8_t *startAddr = packet->payloadCache +
		sizeof(uint8_t)*packet->cachedSize;

	// would adding another payload go OOB?
	// -1 because we need packet term as well
	if (startAddr > &packet->data[MUP_MAX_PAYLOAD_SIZE - 1]) {

		status->error = MUP_ERR_OOB;
		return;
	}

	paylen = mup_payload_len(payload);

	// payload/packet would not have been terminated properly
	if (paylen >= MUP_MAX_PAYLOAD_SIZE) {

		status->error = MUP_ERR_BAD_PACK_LEN;
		return;
	}

	packet->payloadCache = startAddr;

	// +1 to include the terminator
	for (int i = 0; i < paylen + 1; i++) {

		packet->payloadCache[i] = payload[i];
	}

#ifdef MUP_CFG_CONV_STRINGS
	packet->payloadCache[paylen] = MUP_PAYLOAD_TERM;
#endif

	packet->payloadCount++;
	packet->cachedPayload++;
	// paylen incremented to allow for terminator
	packet->cachedSize = ++paylen;

	packet->size += paylen;
	status->size = paylen;
}

// WARNING: this function returns NULL if the payload wasn't found.
// make sure to check its return value!
uint8_t *mup_yield_payload(struct MUPacket *packet, int number) {

	int payloadCount = 0;
	int idx = MUP_HEADER_SIZE_W_CRC;
	uint8_t *start = &packet->data[MUP_HEADER_SIZE_W_CRC];

	if (number == packet->cachedPayload) {

		return packet->payloadCache;
	}

	if (number == packet->cachedPayload + 1) {

		start = packet->payloadCache + sizeof(uint8_t)*packet->cachedSize;
		goto MUP_YIELD_SET_CACHE;
	}

	for (; idx < MUP_MAX_SIZE && payloadCount != number;
		 payloadCount += (packet->data[idx] == MUP_PAYLOAD_TERM), idx++);

	if (idx == MUP_MAX_SIZE) {

		// packet scan returned no payload with that index
		return NULL;
	}

	start = packet->data + sizeof(uint8_t)*idx;

	MUP_YIELD_SET_CACHE:

	packet->payloadCache = start;
	packet->cachedSize = mup_payload_len(packet->payloadCache) + 1;
	packet->cachedPayload = number;

	return start;
}

void mup_serialize(struct MUPacket *packet,
		struct MUStatus *status) {

	int validPayloadCount;

	mup_status_factory(status);

#ifdef MUP_ENFORCE_TYPE
	if (packet->type == MUP_TYPE_UNSET ||
			packet->type > MUP_MAX_TYPE) {

		status->error = MUP_ERR_BAD_TYPE;
		return;
	}
#endif

#ifdef MUP_ENFORCE_ARG
	if (packet->arg > ArgumentBounds[packet->type]) {

		status->error = MUP_ERR_BAD_ARG;
		return;
	}
#endif

	validPayloadCount = PayloadCounts[packet->type];

	packet->data[MUP_HEADER_TYPE_IDX] = packet->type;
	packet->data[MUP_HEADER_ARG_IDX] = packet->arg;
	packet->data[MUP_HEADER_VERSION_IDX] = packet->version;

	if (validPayloadCount == 0) {

		packet->data[MUP_HEADER_SIZE_WO_CRC] = MUP_PACKET_TERM;
		status->size = MUP_HEADER_SIZE_WO_CRC + 1;

		packet->data[MUP_HEADER_SIZE_START_IDX] = status->size << 8;
		packet->data[MUP_HEADER_SIZE_END_IDX] = status->size & 0xFF;
		return;
	}

#ifdef MUP_ENFORCE_PAYL_COUNT
	if (packet->payloadCount < validPayloadCount) {

		status->error = MUP_ERR_BAD_PAYL_COUNT;
		return;
	}
#endif

	// crc in mup is calc'd on the entire packet including header
	// with the crc field zero'd out
	for (int i = MUP_HEADER_CRC_START_IDX; i < MUP_HEADER_SIZE_W_CRC; i++) {

		packet->data[i] = 0x00;
	}

	packet->size += MUP_HEADER_SIZE_W_CRC;
	packet->data[packet->size] = MUP_PACKET_TERM;
	status->size = ++packet->size;

	packet->data[MUP_HEADER_SIZE_START_IDX] = status->size >> 8;
	packet->data[MUP_HEADER_SIZE_END_IDX] = status->size & 0xFF;

	packet->crc32 = mup_calc_crc(packet->data, status->size);

	packet->data[MUP_HEADER_CRC_START_IDX] = packet->crc32 >> 24;
	packet->data[MUP_HEADER_CRC_START_IDX + 1] = packet->crc32 >> 16;
	packet->data[MUP_HEADER_CRC_START_IDX + 2] = packet->crc32 >> 8;
	packet->data[MUP_HEADER_CRC_START_IDX + 3] = packet->crc32 & 0xFF;
}

void mup_deserialize(struct MUPacket *packet, struct MUStatus *status) {

	uint32_t calcedCrc; // calc is short for calculator btw im just usin slang
	int validPayloadCount;
#ifdef MUP_ENFORCE_PAYL_COUNT
	int runningPayloadCount = 0;
#endif

	mup_status_factory(status);
#ifdef MUP_CFG_BYOB
	mup_packet_factory(packet, NULL, false);
#else
	mup_packet_factory(packet, false);
#endif

	packet->version = packet->data[MUP_HEADER_VERSION_IDX];

#ifdef MUP_ENFORCE_VERSION
	if (packet->version != MUP_PROTO_VERSION) {

		status->error = MUP_ERR_BAD_VERSION;
		return;
	}
#endif

	packet->type = packet->data[MUP_HEADER_TYPE_IDX];

#ifdef MUP_ENFORCE_TYPE
	if (packet->type > MUP_MAX_TYPE) {

		status->error = MUP_ERR_BAD_TYPE;
		return;
	}
#endif

	packet->arg = packet->data[MUP_HEADER_ARG_IDX];

#ifdef MUP_ENFORCE_ARG
	if (packet->arg > ArgumentBounds[packet->type]) {

		status->error = MUP_ERR_BAD_ARG;
		return;
	}
#endif

	packet->size = packet->data[MUP_HEADER_SIZE_START_IDX] << 8 |
		packet->data[MUP_HEADER_SIZE_END_IDX];

#ifdef MUP_ENFORCE_SIZE
	if (packet->size > MUP_MAX_SIZE) {

		status->error = MUP_ERR_BAD_PACK_LEN;
		return;
	}
#endif

	// could be dangerous if you dont enforce types
	validPayloadCount = PayloadCounts[packet->type];

	if (validPayloadCount == 0) {

#ifdef MUP_ENFORCE_SIZE
		status->error = mup_packet_len(packet->data) + 1 == packet->size ?
			MUP_ERR_NONE : MUP_ERR_BAD_PACK_LEN;
#endif
		return;
	}

	packet->crc32 = packet->data[MUP_HEADER_CRC_START_IDX] << 24;
	packet->crc32 |= packet->data[MUP_HEADER_CRC_START_IDX + 1] << 16;
	packet->crc32 |= packet->data[MUP_HEADER_CRC_START_IDX + 2] << 8;
	packet->crc32 |= packet->data[MUP_HEADER_CRC_START_IDX + 3];

	// in MUP crc is calced on the entire packet with crc field blanked out
	for (int i = MUP_HEADER_CRC_START_IDX; i < MUP_HEADER_SIZE_W_CRC; i++) {

		packet->data[i] = 0x00;
	}

	calcedCrc = mup_calc_crc(packet->data, packet->size);
	if (calcedCrc != packet->crc32) {

		status->error = MUP_ERR_BAD_CRC;
		return;
	}

	// put it back for integrity's sake
	packet->data[MUP_HEADER_CRC_START_IDX] = packet->crc32 >> 24;
	packet->data[MUP_HEADER_CRC_START_IDX + 1] = packet->crc32 >> 16;
	packet->data[MUP_HEADER_CRC_START_IDX + 2] = packet->crc32 >> 8;
	packet->data[MUP_HEADER_CRC_START_IDX + 3] = packet->crc32 & 0xFF;

	// NOTE: this check still happens even if MUP_ENFORCE_SIZE isn't set
	// because if there is no terminator it isn't a valid packet.
	// if the terminator isn't where it says it is there might be corruption.
	if (packet->size == 0 || packet->data[packet->size - 1] != MUP_PACKET_TERM) {

		status->error = MUP_ERR_BAD_PACK_LEN;
		return;
	}

	status->size = packet->size;

#ifdef MUP_ENFORCE_PAYL_COUNT

	for (int i = 0; i < MUP_MAX_SIZE;
		 runningPayloadCount += (packet->data[i] == MUP_PAYLOAD_TERM), i++);

	if (runningPayloadCount < PayloadCounts[packet->type]) {

		status->error = MUP_ERR_BAD_PAYL_COUNT;
		return;
	}
#endif
}

void mup_construct_api_msg(struct MUPacket *packet, uint8_t arg, struct MUStatus *status) {

	mup_status_factory(status);
#ifdef MUP_CFG_BYOB
	mup_packet_factory(packet, NULL, false);
#else
	mup_packet_factory(packet, false);
#endif

	packet->type = MUP_TYPE_API;
	packet->arg = arg;

	mup_serialize(packet, status);
}

void mup_construct_bad_msg(struct MUPacket *packet, struct MUStatus *status) {

#ifdef MUP_CFG_BYOB
	mup_packet_factory(packet, NULL, false);
#else
	mup_packet_factory(packet, false);
#endif
	packet->type = MUP_TYPE_BAD;

	switch (status->error) {

		case MUP_ERR_OOB:
			packet->arg = MUP_ARG_BAD_PACK_LEN;
			break;
		case MUP_ERR_BAD_CRC:
			packet->arg = MUP_ARG_BAD_CRC;
			break;
		case MUP_ERR_BAD_VERSION:
			packet->arg = MUP_ARG_BAD_VERSION;
			break;
		case MUP_ERR_BAD_TYPE:
			packet->arg = MUP_ARG_BAD_TYPE;
			break;
		case MUP_ERR_BAD_ARG:
			packet->arg = MUP_ARG_BAD_ARG;
			break;
		case MUP_ERR_BAD_PACK_LEN:
			packet->arg = MUP_ARG_BAD_PACK_LEN;
			break;
		case MUP_ERR_BAD_PAYL_COUNT:
			packet->arg = MUP_ARG_BAD_PAYL_COUNT;
			break;
		case MUP_ERR_BAD_PAYL_VALUE:
			packet->arg = MUP_ARG_BAD_PAYL;
			break;
		case MUP_ERR_NONE:
		default:
			packet->arg = MUP_ARG_BAD_INTERN_ERR;
	}

	mup_serialize(packet, status);
}

void mup_process_events(struct MUPacket *rxPacket, struct MUPacket *txPacket,
		struct MUStatus *status, bool doDeserial) {

	mup_status_factory(status);

	enum MUPError storeErr;
	enum MUPEvent storeEvt;

	if (doDeserial) {

		mup_deserialize(rxPacket, status);
		if (status->error != MUP_ERR_NONE) {

			status->event = status->error == MUP_ERR_BAD_CRC ?
				MUP_EVT_CORRUPTED : MUP_EVT_YOUR_BAD;

			storeErr = status->error;
			storeEvt = status->event;

			mup_construct_bad_msg(txPacket, status);

			status->error = storeErr;
			status->event = storeEvt;
			status->doReply = true;
			return;
		}
	}

	switch (rxPacket->type) {

		case MUP_TYPE_INIT:

			if (rxPacket->arg == MUP_ARG_INIT_ACCEPT ||
					rxPacket->arg == MUP_ARG_INIT_INIT) {

				status->event = MUP_EVT_INIT;
				break;
			}

			status->event = MUP_EVT_INTERN_ERR;
			break;
		case MUP_TYPE_BAD:

			status->event = MUP_EVT_MY_BAD;
			break;
		case MUP_TYPE_CONTERM:

			status->event = MUP_EVT_CONTERM;
			break;
		case MUP_TYPE_SONAR:

			if (rxPacket->arg == MUP_ARG_SONAR_PONG) {

				status->event = MUP_EVT_SONAR_PONG;
				break;
			}

#ifdef MUP_CFG_BYOB
			mup_packet_factory(txPacket, NULL, false);
#else
			mup_packet_factory(txPacket, false);
#endif
			txPacket->type = MUP_TYPE_SONAR;
			txPacket->arg = MUP_ARG_SONAR_PONG;
			mup_serialize(txPacket, status);

			status->event = MUP_EVT_SONAR_PING;
			status->doReply = true;
			break;
		case MUP_TYPE_SEND:

			status->event = MUP_EVT_RECV;
			break;
		case MUP_TYPE_API:

			if (rxPacket->arg == MUP_ARG_API_SUCCESS) {

				status->event = MUP_EVT_API_SUCCESS;
				return;
			}

			status->event = MUP_EVT_API_ERR;
			break;

		default:
			status->error = MUP_ERR_INTERNAL;
			status->event = MUP_EVT_INTERN_ERR;

			storeErr = status->error;
			storeEvt = status->event;

			mup_construct_api_msg(txPacket, MUP_ARG_BAD_INTERN_ERR, status);

			status->error = storeErr;
			status->event = storeEvt;
			status->doReply = true;
	}
}
