/* libmup.h - M.A.R-T>Y. Micro Protocol (MuP) is an endpoint-based request/response RPC
 * protocol for small-ish interlinked systems.

 * The protocol spec can be found at https://hoff.industries/release-mup.html
 * Last updated 06/08/2024

 * please report bugs to theworldisyours@mailbox.org

 * This library licensed with BSD 3-Clause:

Copyright 2023-2024 hoff.industries

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#define H_LIBMUP

/* libmup #define modifiers:

 * MUP_CFG_CONV_STRINGS - treat '\0' like the MuP payload
	terminator 0xFE, and convert automatically on serialization

 * MUP_CFG_BYOB - 'bring your own buffer' i.e. initialize packets
	with your own buffer instead of having the library alloc one.
	make sure it's geq MUP_MAX_SIZE or change the max size define.

 * MUP_CFG_BYOC - 'bring your own crc' i.e. use your own crc function
	instead of the table and function this library provides. You might
	want this if you have hardware accelerated CRC or an alternative
	method of calculation - WARNING: use mup_register_crc_func to set the
	calculation function pointer or you'll use the dummy constant function!

 * MUP_ENFORCE_VERSION - reject packets whose version field
	doesn't match the one at compile time

 * MUP_ENFORCE_TYPE - bounds check types and reject if bad
 * MUP_ENFORCE_ARG - bounds check args and reject if bad
 * MUP_ENFORCE_SIZE - check reported size against actual size and reject if bad
 * MUP_ENFORCE_PAYL_COUNT - verify payload counts are correct for each type

 If you're confident you'll never send a bad message you can undefine the
	enforcements to speed up de/serialization.

 * MUP_CFG_DEBUG - turn on all enforcements. */

#define MUP_CFG_CONV_STRINGS
#define MUP_CFG_DEBUG

// NOTE: add/remove your #define modifiers *ABOVE* this line.

#ifdef MUP_CFG_DEBUG
#define MUP_ENFORCE_VERSION
#define MUP_ENFORCE_TYPE
#define MUP_ENFORCE_ARG
#define MUP_ENFORCE_SIZE
#define MUP_ENFORCE_PAYL_COUNT
#endif

#include <stdint.h>
#include <stdbool.h>

#define MUP_MAX_SIZE 2048

#define MUP_PAYLOAD_TERM 0xFE
#define MUP_PACKET_TERM 0xFF

#define MUP_TYPE_UNSET		0x00
#define MUP_TYPE_INIT		0x01
#define MUP_TYPE_BAD		0x02
#define MUP_TYPE_CONTERM	0x03
#define MUP_TYPE_SONAR		0x04
#define MUP_TYPE_SEND		0x05
#define MUP_TYPE_API		0x06

#define MUP_MAX_TYPE 6

#define MUP_PAYL_COUNT_UNSET	0
#define MUP_PAYL_COUNT_INIT		1
#define MUP_PAYL_COUNT_BAD		0
#define MUP_PAYL_COUNT_CONTERM	0
#define MUP_PAYL_COUNT_SONAR	0
#define MUP_PAYL_COUNT_SEND		1
#define MUP_PAYL_COUNT_API		0

#define MUP_ARG_UNSET_UNSET			0x00
#define MUP_ARG_UNSET_MAX			0x00

#define MUP_ARG_INIT_INIT			0x01
#define MUP_ARG_INIT_ACCEPT			0x02
#define MUP_ARG_INIT_REJECT			0x03
#define MUP_ARG_INIT_MAX			0x03

#define MUP_ARG_BAD_TYPE			0x01
#define MUP_ARG_BAD_PACK_LEN		0x02
#define MUP_ARG_BAD_ARG				0x03
#define MUP_ARG_BAD_PAYL			0x04
#define MUP_ARG_BAD_PAYL_COUNT		0x05
#define MUP_ARG_BAD_VERSION			0x0A
#define MUP_ARG_BAD_CRC				0x0B
#define MUP_ARG_BAD_INTERN_ERR		0x0C
#define MUP_ARG_BAD_MAX				0x0C

#define MUP_ARG_CONTERM_CLEAN		0x01
#define MUP_ARG_CONTERM_SPAM		0x02
#define MUP_ARG_CONTERM_MANY_BAD	0x03
#define MUP_ARG_CONTERM_MAX			0x03

#define MUP_ARG_SONAR_PING			0x00
#define MUP_ARG_SONAR_PONG			0x01
#define MUP_ARG_SONAR_MAX			0x01

#define MUP_ARG_SEND_DIRECT			0x01
#define MUP_ARG_SEND_MAX			0x01

#define MUP_ARG_API_SUCCESS			0x01
#define MUP_ARG_API_ERR_BUSY		0x03
#define MUP_ARG_API_ERR_RANGE		0x04
#define MUP_ARG_API_ERR_ENDPOINT	0x05
#define MUP_ARG_API_ERR_VALUE		0x06
#define MUP_ARG_API_ERR_COUNT		0x07
#define MUP_ARG_API_ERR_EXEC		0x0A
#define MUP_ARG_API_MAX				0x0A

#define MUP_HEADER_SIZE_W_CRC		9
#define MUP_HEADER_SIZE_WO_CRC		5

#define MUP_HEADER_TYPE_IDX			0
#define MUP_HEADER_ARG_IDX			1
#define MUP_HEADER_VERSION_IDX		2

#define MUP_HEADER_SIZE_START_IDX	3
#define MUP_HEADER_SIZE_END_IDX		4

#define MUP_HEADER_CRC_START_IDX	5

#define MUP_SEND_PAYL_ENDPOINT_IDX	0

// this is probably off by one - TODO: confirm
#define MUP_MAX_PAYLOAD_SIZE		MUP_MAX_SIZE - MUP_HEADER_SIZE_W_CRC - 3

const extern int MUP_LIB_VERSION[];
const extern int MUP_PROTO_VERSION;
const extern char MUP_AUTHOR[];
const extern char MARTY_FLAVOUR_TEXT[];

extern uint32_t (*mup_calc_crc)(const void *buf, int size);

enum MUPError {

	MUP_ERR_NONE,

	MUP_ERR_OOB,
	MUP_ERR_INTERNAL,

	MUP_ERR_BAD_VERSION,
	MUP_ERR_BAD_CRC,
	MUP_ERR_BAD_TYPE,
	MUP_ERR_BAD_ARG,
	MUP_ERR_BAD_PACK_LEN,
	MUP_ERR_BAD_PAYL_COUNT,
	MUP_ERR_BAD_PAYL_VALUE,
};

enum MUPEvent {

	MUP_EVT_NONE,
	MUP_EVT_INIT,
	MUP_EVT_INTERN_ERR,
	MUP_EVT_API_ERR,
	MUP_EVT_API_SUCCESS,
	MUP_EVT_RECV,
	MUP_EVT_MY_BAD,
	MUP_EVT_YOUR_BAD,
	MUP_EVT_CORRUPTED,
	MUP_EVT_SONAR_PING,
	MUP_EVT_SONAR_PONG,
	MUP_EVT_CONTERM
};

struct MUPacket {

	uint8_t type;
	uint8_t arg;
	uint8_t version;
	uint16_t size;
	uint32_t crc32;

	int payloadCount;

	uint8_t *payloadCache;
	int cachedPayload;
	int cachedSize;

#ifdef MUP_CFG_BYOB
	uint8_t *data;
#else
	uint8_t data[MUP_MAX_SIZE];
#endif
};

struct MUStatus {

	int size;
	enum MUPError error;
	enum MUPEvent event;
	bool doReply;
};

struct MUEndpoint {
	int endpoint;
	char *name;

	// define the min and
	// max number of payloads
	// for the endpoint
	int min;
	int max;

	// only accept ascii chars?
	bool isAscii;
};

void mup_register_crc_func(uint32_t (*crc_calc_func)(const void *buf, int size));

#ifndef MUP_CFG_BYOC
uint32_t mup_crc32(const void *buf, int size);
#endif

#ifdef MUP_CFG_BYOB
void mup_packet_factory(struct MUPacket *packet, uint8_t *buffer, bool zero);
#else
void mup_packet_factory(struct MUPacket *packet, bool zero);
#endif
void mup_status_factory(struct MUStatus *status);
void mup_endpoint_factory(struct MUEndpoint *endpoint);

int mup_payload_len(uint8_t *payload);
int mup_packet_len(uint8_t *packet);

void mup_add_payload(struct MUPacket *packet,
		uint8_t *payload, struct MUStatus *status);
uint8_t *mup_yield_payload(struct MUPacket *packet, int number);

void mup_serialize(struct MUPacket *packet, struct MUStatus *status);
void mup_deserialize(struct MUPacket *packet, struct MUStatus *status);

void mup_construct_bad_msg(struct MUPacket *packet, struct MUStatus *status);
void mup_construct_api_msg(struct MUPacket *packet, uint8_t arg, struct MUStatus *status);

void mup_process_events(struct MUPacket *rxPacket, struct MUPacket *txPacket,
	struct MUStatus *status, bool doDeserial);
